"""
This module contains classes to represent issuers fo audio samples.
An issuer represents the person/machine/thing, that is creator of an audio
segment.
"""

from .issuer import Issuer  # noqa: F401

from .speaker import Speaker  # noqa: F401
from .speaker import Gender  # noqa: F401
from .speaker import AgeGroup  # noqa: F401

from .artist import Artist  # noqa: F401
